#include <LEDA/graph.h>
#include <LEDA/d_array.h>
#include <LEDA/h_array.h>
#include <LEDA/map.h>
#include <LEDA/list.h>
#include <LEDA/dictionary.h>

#ifndef GA_H
#define GA_H

//
// GraphEdAttrType
//

enum GraphEdAttrType {
    GraphEdAttrUndefined,
    GraphEdAttrInt,
    GraphEdAttrString,
    GraphEdAttrList
};


//
// Forward Declaration
//

class GAList;
class GA;


//
// Global Key Mapper
//

typedef int KeyType;

class KeyMapper {
    dictionary<string,KeyType> keys;
    dictionary<KeyType,string> keys_inverse;
    int count;

public:

    KeyMapper();
    ~KeyMapper();

    KeyType  add (const string& s);
    KeyType  add (const string& s, KeyType key);
    KeyType  key (const string& s) const;
    string   definition (const KeyType key);
    bool     defined (const string& s) const;
};


extern KeyMapper keymapper;


//
// GAValue
//

class GAValue {

    GraphEdAttrType type;

    struct {
	int     integer;
	string  str;
	GAList  *sublist;
    } data;

public:

    GAValue ();
    GAValue (int i);
    GAValue (const string& s);
    GAValue (GAList* l);

    ~GAValue ();

    GraphEdAttrType get_type () const;
    int             get_int() const ;
    const string&   get_string() const ;
    GAList*         get_list() const;
};

ostream& operator<< (ostream &out, const GAValue* v);



//
// GAValueList
//

typedef list<GAValue*> GAValueList;
ostream& operator<< (ostream &out, const GAValueList& values);


//
// GA
//

class GA : public GAValueList {

  KeyType     thekey;

public:
  
  GA ();
  GA (const KeyType key);

  ~GA ();

  void put_key (const KeyType key);
  
  const KeyType key () const;
};

ostream& operator<< (ostream &out, const GA* a);



//
// GAList
//

class GAList : public list<GA*> {

public:

    GAList ();
    ~GAList ();

    GA* find (const KeyType key);

    friend ostream& operator<< (ostream& out, const GAList* attrs);
};

ostream& write_quoted (ostream &out, const string& text);


//
// GAGraph;
//

typedef node_array<GAList*> GANodeAttrs;
typedef edge_array<GAList*> GAEdgeAttrs;
typedef GAList*             GAGraphAttrs;


class GAGraph {

    graph*          the_graph;

    GANodeAttrs     the_node_attrs;
    GAEdgeAttrs     the_edge_attrs;
    GAGraphAttrs    the_graph_attrs;

    node_array<int> the_numbers;

public:

    GAGraph ();
    GAGraph (graph* g);
    ~GAGraph ();

    graph* get () const;
    void   set (graph*g);

    GAList*& node_attrs(node v);
    GAList*& edge_attrs(edge e);
    GAList*& graph_attrs();

    int  number(node v) const;
    void set_number(node v, int n);

};

ostream& operator< (ostream& out, GAGraph& G);
ostream& operator<< (ostream& out, GAGraph& G);


//
// KeyMapper
//

inline KeyMapper::~KeyMapper() {
    ;
};


inline KeyType KeyMapper::key (const string& s) const {
    return keys.access(s);
};

inline bool KeyMapper::defined (const string& s) const {
    return (keys.lookup(s) != nil);
};

inline string KeyMapper::definition (const KeyType key) {
    return keys_inverse.access(key);
}

//
// GAValue
//

inline GAValue::GAValue () {
    type = GraphEdAttrUndefined;
};

inline GAValue::GAValue (int i) {
    type = GraphEdAttrInt;
    data.integer = i;
};

inline GAValue::GAValue (const string& s) {
    type = GraphEdAttrString;
    data.str = s;
};

inline GAValue::GAValue (GAList* l) {
    type = GraphEdAttrList;
    data.sublist = l;
};

inline GAValue::~GAValue () {
    ;
};

inline GraphEdAttrType GAValue::get_type () const {
    return type;
};

inline int GAValue::get_int() const {
    return data.integer;
};

inline const string& GAValue::get_string() const {
    return data.str;
};
 
inline GAList* GAValue::get_list() const {
    return data.sublist;
};



//
// GA
//


inline GA::GA () {
    ;
};

inline GA::GA (const KeyType key) {
    thekey = key;
};

inline GA::~GA () {
    ;
};


inline void GA::put_key (const KeyType key) {
    thekey = key;
};


inline const KeyType GA::key () const {
    return thekey;
};

// inline const GAValueList& GA::values () const {
//     return thevalues;
// };



//
// GAList
//

inline GAList::GAList () { };
inline GAList::~GAList () { };

inline GA* GAList::find (const KeyType key) {
    list_item a;
    forall_items (a, *this) { if (inf(a)->key() == key) return inf(a); }
    return NULL; /* Tim */
};


//
// GAGraph
//

inline graph* GAGraph::get() const {
    return the_graph;
}

inline GAList*& GAGraph::graph_attrs() {
    return the_graph_attrs;
}

inline GAList*& GAGraph::node_attrs(node v) {
    return the_node_attrs[v];
}

inline GAList*& GAGraph::edge_attrs(edge e) {
    return the_edge_attrs[e];
}


inline int GAGraph::number (node v) const {
    return the_numbers[v];
}

inline void GAGraph::set_number (node v, int n) {
    the_numbers[v] = n;
}

#endif
