#include <Tim/GA.h>

//
// KeyMapper
//

KeyMapper::KeyMapper() {
    count = 0;

    add("NodePosition:",        add("NP"));
    add("Label:",               add("LABEL:"));
    add("Nr:",                  add("NR:"));

    add ("Font:");
    add ("NodeType:");
    add ("EdgeType:");
    add ("NodeStyle:",          add ("NSTYL"));
    add ("EdgeStyle:",          add ("ESTYL"));

    add("Directed:",            add("DIR"));
    add("GraGra:",              add("GRAGRA"));
    add("GlobalEmbed:",         add("GEMBED"));
    add("EmbeddingMatch:",      add("EMBMATCH"));

    add("WorkingAreaSize:",     add("WA"));
    add("WorkingAreaWindow:",   add("WAWIN"));
    add("ScrollOffset:",        add("SC"));
    add("GridWidth:",           add("GR"));

    add("NodeSize:",            add("NS"));
    add("NodeTypeIndex:",       add("NTI"));
    add("NodeFontIndex:",       add("NFI"));
    add("NodeLabelPlacement:",  add("NLP"));
    add("NodeEdgeInterface:",   add("NEI"));
    add("NodeLabelVisibility:", add("NLV"));
    add("NodePosition:",        add("NP"));

    add("EdgeTypeIndex:",       add("ETI"));
    add("EdgeFontIndex:",       add("EFI"));
    add("EdgeLabelVisibility:", add("ELV"));
    add("ArrowLength:",         add("AL"));
    add("ArrowAngle:",          add("AA"));
    add("EdgeLine:" ,           add("EL"));
    add("EdgeLabelSize:",       add("ELS"));

    add("Color:",               add("COL"));
};


KeyType KeyMapper::add (const string& s) {
    dic_item it = keys.lookup(s);
    if (it != NULL) {
	return keys.inf(it);
    } else {
	count ++;
	keys_inverse.insert(count, s);
	return keys.inf(keys.insert(s, count));
    }
};

KeyType KeyMapper::add (const string& s, KeyType key) {
    dic_item it = keys.lookup(s);
    if (it != NULL) {
	return keys.inf(it);
    } else {
	keys_inverse.insert(key, s);
	return keys.inf(keys.insert(s, key));
    }
};



//
// Helpers
//

ostream& write_quoted (ostream &out, const string& text)
{
    /* Taken from GraphEd 4.0 source code */

    int i, length;
    const char ESCAPE_CHARACTER = '\\';

    out << '"';
    length = text.length();
    for (i = 0; i<length; i++) {
	switch (text[i]) {
	case '"' :
	case '{' : /* GraphEd Compatibility */
	case '}' : /* GraphEd Compatibility */
	    out << ESCAPE_CHARACTER << text[i];
	    break;
	case '\015' :
	case '\n' :
	    out << ESCAPE_CHARACTER << 'n';
	    break;
	case '\t' :
	    out << ESCAPE_CHARACTER << 't';
	    break;
	case ESCAPE_CHARACTER :
	    out << ESCAPE_CHARACTER << ESCAPE_CHARACTER;
	    break;
	default :
	    out << text[i];
	    break;
	}
    }
    return out << '"';
}


//
// GAValue
//


ostream& operator<< (ostream &out, const GAValue* v) {

    if (v != NULL) {
	switch (v->get_type()) {
	case GraphEdAttrInt:
	    return out << v->get_int();
	    break;
	case GraphEdAttrString:
	    return write_quoted(out, v->get_string());
	    break;
	case GraphEdAttrList:
	    return out << "[\n" << v->get_list() << "]\n";
	    break;
	case GraphEdAttrUndefined: /* Tim */
	    ; /* Tim */
	}
    }

    return out;
}


//
// GA
//

ostream& operator<< (ostream &out, const GA* a)
{
  string definition;
  list_item value;

  if (a != NULL) {
      definition = keymapper.definition(a->key());
      if (definition[0] != '.') {
	  out << definition;
	  forall_items (value, *a) {
	      out << ' ' << (*a)[value];
	  }      
      }
  }
  
  return out;
}


//
// GAList
//

ostream& operator<< (ostream& out, const GAList* attrs)
{
    string s;
    list_item a;

    if (attrs != NULL) {
	forall_items (a, *attrs) {
	    out << ' ' << (*attrs)[a];
	}
    }    
    return out;
}


//
//  GAGraph
//

GAGraph::GAGraph () {
    the_graph       = NULL;
    the_graph_attrs = NULL;
}

GAGraph::GAGraph (graph* g) {
    GAGraph();
    set(g);
}

void GAGraph::set (graph* g) {
    the_graph = g;

//  node_map version
//  the_numbers.init (*the_graph,0);
//  the_node_attrs.init (*the_graph, NULL);
//  the_edge_attrs.init (*the_graph, NULL);

//  node_array version
    the_numbers.init(*the_graph,    1000, -1);
    the_node_attrs.init(*the_graph, 1000, NULL);
    the_edge_attrs.init(*the_graph, 1000, NULL);

    the_graph_attrs = 0;
}

GAGraph::~GAGraph () {
    the_graph = 0;
    delete the_graph_attrs;
}

//
// Output (GraphEd 4.* compatible file format)
//

ostream& operator<< (ostream& out, GAGraph& G)
{
  node v;
  edge e;
  graph *g = G.get();

  if (g == NULL) {
     return out;
  }

  KeyType nr_key = keymapper.add("NR:");
  KeyType label_key = keymapper.add("LABEL:");

  // Determine the highest number
  int highestnumber = 0;
  forall_nodes (v, *g) {
      highestnumber = G.number(v) > highestnumber ?
	  G.number(v) :
	  highestnumber;
  }

  // Fill unused numbers
  forall_nodes (v, *g) {
      if (G.number(v) == -1) {
	  G.set_number (v, ++highestnumber);
      }
  }

  out << "GRAPH \"\" = ";
  if (g->is_directed ()) {
     out << "DIRECTED\n";
  } else {
     out << "UNDIRECTED\n";
  }

  // Output graph attributes
  GAList* attrs = G.graph_attrs();
  if (attrs != NULL) {
      GA* label = attrs->find(label_key);
      if (label != NULL) {
	  out << ' ' << label->head();
      }
      out << " {$" << attrs << " $}\n";
  }

  forall_nodes (v, *g) {

      out << ' ' << G.number(v);

      // Output node attributes
      GAList* attrs = G.node_attrs(v);
      if (attrs != NULL) {
	  out << " {$" << attrs << " $} ";
	  GA* label = attrs->find(label_key);
	  if (label != NULL) {
	      out << ' ' << label->head();
	  }
      }
      out << "\n";
      
      forall_adj_edges (e, v) {

	  out << "  " << G.number(g->target(e));

	  // Output node attributes
	  GAList* attrs = G.edge_attrs(e);
	  if (attrs != NULL) {
	      out << " {$" << attrs << " $} ";
	      GA* label = attrs->find(label_key);
	      if (label != NULL) {
		  out << ' ' << label->head();
	      }
	  }
	  out << '\n' ;
      }
      out << ";\n";
  }
  return out << "END\n";
}
