#include <ostream.h>
#include <LEDA/queue.h>
#include <Tim/graph.h>
#include <Tim/degree_sequence.h>

/*****************************************************************************/
/* Degree sequence                                                           */
/* Stored in non-decreasing order in indices 0..(n-1)                        */
/*****************************************************************************/

bool operator == (degree_sequence &x, degree_sequence &y) {
	int xi = x.high ();
	int yi = y.high ();

	while ((xi >= x.low ()) && (yi >= y.low ())) {
		if (x[xi] != y[yi]) {
			return false;
		}
		xi--;
		yi--;
	}
#if 0
	if (((xi >= x.low ()) && (x[xi] > 0)) ||
	    ((yi >= y.low ()) && (y[yi] > 0))) {
		return false;
	}
#endif
	if ((xi >= x.low ()) || (yi >= y.low ())) {
		return false;
	}
	return true;
}

degree_sequence::degree_sequence (graph &G) :
	array<degree_t> (G.number_of_nodes ())
{
	node v;
	int i;

	number_of_edges = G.number_of_edges ();
	lower = 0;
	upper = G.number_of_nodes () - 1;

	i = low ();
	forall_nodes (v, G) {
		(*this)[i] = G.degree (v);
		i++;
	}
	sort ();
}

#define TERSE_PRINT_SEQUENCE

ostream & operator << (ostream &out, degree_sequence &d) {
	int i;
#ifdef TERSE_PRINT_SEQUENCE
	int occurrences;
#endif
	out << "(";
	if (d.low () <= d.high ()) {
		for (i = d.low ();; i++) {
			out << d[i];
#ifdef TERSE_PRINT_SEQUENCE
			occurrences = 1;
			while ((i < d.high ()) && (d[i] == d[i+1])) {
				i++;
				occurrences++;
			}
			if (occurrences > 1) {
				out << "^" << occurrences;
			}
#endif
			if (i == d.high ()) {
				break;
			}
			out << ",";
		}
	}
	return out << ")";
}

/*****************************************************************************/
/* Generation in lexicographic order                                         */
/*****************************************************************************/

bool degree_sequence::lexicographic_first () {
	int i;
	degree_t remainder = excess;
	degree_t delta = upper - lower;
	for (i = low (); i <= high (); i++) {
		(*this)[i] = lower;
	}
	if (remainder < 0) {
		return false;
	}
	for (i = high (); i >= low (); i--) {
		if (remainder > delta) {
			(*this)[i] += delta;
			remainder  -= delta;
		} else {
			(*this)[i] += remainder;
			return true;
		}
	}
	return false;
}

bool degree_sequence::lexicographic_next () {
	int i, j;
	degree_t maxdegree = (*this)[high()];
	degree_t remainder = maxdegree;
	for (i = high () - 1; i >= low (); i--) {
		remainder += (*this)[i];
		if ((maxdegree - (*this)[i]) >= 2) {
			break;
		}
	}
	if (i >= low ()) {
		degree_t newdeg = (*this)[i] + 1;
		for (j = i; j <= high (); j++) {
			remainder -= (*this)[j] = newdeg;
		}
		for (j = high (); remainder > 0; j--) {
			if (((*this)[j] + remainder) > upper) {
				remainder -= upper - (*this)[j];
				(*this)[j] = upper;
			} else {
				(*this)[j] += remainder;
				remainder = 0;
			}
		}
		return true;
	}
	return false;
}

/*****************************************************************************/
/* Checks whether degree sequence is graphical                               */
/* Run Time: O(n^2)                                                          */
/*****************************************************************************/

bool degree_sequence::is_graphical () {
	int r, i, j;
	int lhs = 0;
	int rhs;

	for (i = high (), r = 1; i >= low (); i--, r++) {
		lhs += (*this)[i];
		rhs  = r*(r-1);
		for (j = i-1; j >= low (); j--) {
			rhs += Min ((*this)[j], r);
		} 
		if (lhs > rhs) {
			return false;
		}
	}
	return true;
}

/*****************************************************************************/
/* Checks whether degree sequence is 2-tree graphical                        */
/* Run Time: O(n^2) (Mostly to check whether sequence is graphical at all)   */
/*****************************************************************************/

bool degree_sequence::is_ktree_graphical (int k) {
	int i, n, r, K, D4, D2;
	degree_t m;
	bool alleven;
	degree_t deg;

	if (!is_graphical ()) {
		return false;
	}

	if (k != 2) {
		return false;
	}

	// Must allow for an edge!

	D2 = 0;
	D4 = 0;
	K  = 0;
	m  = 0;
	n  = high () - low () + 1;
	alleven = true;
	for (i = low (); i <= high (); i++) {
		deg = (*this)[i];
		m += deg;
		if (deg % 2) {
			alleven = false;
		}
		if (deg < 2) {
			return false;
		}
		if (deg == 2) {
			D2++;
		} else if (deg == 4) {
			D4++;
		} else if (deg > 4) {
			K++;
		}
	}
	m >>= 1;
	if ((m != 2*n - 3) ||
	    (D2 < 2)  ||
	    (alleven && (D4 > 2*D2-K-3))) {
		return false;
	}
	r = (D2 - 5) >> 1;
	if ((D2 == 5 + 2*r) &&
	    (K == 4) &&
	    ((*this)[high ()] == 5+r) &&
	    ((*this)[high ()-3] == 5+r)) {
		return false;
	}
	return true;
}

/*****************************************************************************/
/* Construct a k-tree from a degree sequence                                 */
/* Run Time:                                                                 */
/*****************************************************************************/

static int K, D3, D4;
static list<int> degrees;
static node_array<bool> pivoted;
static node_array<edge> added_to;
static node_array<edge> back_edge;
static node_array<int>  dp;
static edge_array<int>  leaves;
static node_array<int>  single_fours;
static node_array<int>  double_fours;
static edge_array<int>  split_fours;
static node_array<int>  threes;

void add_leaf (graph &G, const edge &e, int qty = 1) {
	leaves[e] += qty;
	dp[G.source (e)] -= qty;
	dp[G.target (e)] -= qty;
}

void add_leaf (graph &G, const node &u, const node &v, int qty = 1) {
	edge e;
	find_edge (G, e, u, v);
	add_leaf (G, e, qty);
}

void remove_leaf (graph &G, const edge &e) {
	leaves[e]--;
	dp[G.source (e)]++;
	dp[G.target (e)]++;
}

void post_process_leaves (graph &G) {
	edge e;
	node leaf;
	forall_edges (e, G) {
		while (leaves[e] > 0) {
			leaf = G.new_node ();
			G.new_edge (leaf, G.source (e));
			G.new_edge (leaf, G.target (e));
			leaves[e]--;
		}
	}
}

bool find_splittable_edge (graph &G, edge &e, const node &v, const node & u) {
	node w, x;
	bool found_edge;
	forall_adj_nodes (w, v) {
		if ((u != w) && find_edge (G, e, u, w)) {
			found_edge = true;
			forall_adj_nodes (x, u) {
				if ((x != v) && is_adjacent (G, x, w)) {
					found_edge = false;
					break;
				}
			}
			if (found_edge) {
				return true;
			}
		}
	}
	return false;
}

bool find_splittable_edge (graph &G, edge &e, const node &v) {
	node u;
	forall_adj_nodes (u, v) {
		if (find_splittable_edge (G, e, v, u)) {
			return true;
		}
	}
	return false;
}

void add_three (graph &G, node &v) {
	threes[v]++;
	pivoted[v] = true;
	dp[v]--;
	D3--;
}

void remove_three (graph &G, node &v) {
	threes[v]--;
	dp[v]++;
	D3++;
}

node bisect (graph &G, const node &v, edge &e) {
	node bisector = G.new_node ();
	edge new_e = G.new_edge (bisector, G.source (e));
	G.new_edge (bisector, G.target (e));
	G.new_edge (bisector, v);
	G.del_edge (e);
	e = new_e;
	return bisector;
}

void trisect (graph &G, const node &v, edge &e, node &trisector1, node &trisector2) {
	trisector1 = G.new_node ();
	trisector2 = G.new_node ();
	edge new_e = G.new_edge (G.source (e), trisector1);
	G.new_edge (trisector1, trisector2);
	G.new_edge (trisector2, G.target (e));
	G.del_edge(e);
	G.new_edge (trisector1, v);
	G.new_edge (trisector2, v);
	e = new_e;
}

void post_process_threes (graph &G) {
	node v;
	edge e;
	forall_nodes (v, G) {
		if (threes[v] > 0) {
			if (!find_splittable_edge (G, e, v)) {
cout << "COULD NOT FIND SPLITTABLE\n";
				continue;
			}
			while (threes[v] > 0) {
				bisect (G, v, e);
				threes[v]--;
			}
		}
	}
}

void add_double_four (graph &G, node &v) {
	double_fours[v]++;
	pivoted[v] = true;
	dp[v] -= 2;
	degrees.Pop (); // Remove 4's from list
	degrees.Pop ();
	D4 -= 2;
}

void add_single_four (graph &G, node &v) {
	single_fours[v]++;
	pivoted[v] = true;
	dp[v] -= 2;
	degrees.Pop (); // Remove 4 from list
	D4--;
}

void add_split_four (graph &G, edge &e) {
	split_fours[e]++;
	degrees.Pop (); // Remove 4 from list
	D4--;
	dp[G.source (e)]--;
	dp[G.target (e)]--;
}

void remove_double_four (graph &G, node &v) {
	double_fours[v]--;
	degrees.append (1);
	degrees.append (1);
	D4 += 2;
	dp[v] += 2;
}

void remove_single_four (graph &G, node &v) {
	single_fours[v]--;
	degrees.append (1);
	D4++;
	dp[v] += 2;
}

void post_process_fours (graph &G) {
	node v;
	edge e;
	edge split;
	forall_nodes (v, G) {
		if ((single_fours[v] > 0) || (double_fours[v] > 0)) {
			if (!find_splittable_edge (G, e, v)) {
cout << "COULD NOT FIND SPLITTABLE\n";
				continue;
			}
			while (single_fours[v] > 0) {
				node bisector = bisect (G, v, e);
				node leaf = G.new_node ();
				G.new_edge (leaf, v);
				G.new_edge (leaf, bisector);
				single_fours[v]--;
			}
			while (double_fours[v] > 0) {
				node trisector1;
				node trisector2;
				trisect (G, v, e, trisector1, trisector2);
				node leaf = G.new_node ();
				G.new_edge (leaf, trisector1);
				G.new_edge (leaf, trisector2);
				double_fours[v]--;
			}
		}
	}
	forall_edges (e, G) {
		if (split_fours[e] > 0) {
			if (find_splittable_edge (G, split, G.source (e), G.target (e))) {
				node bisector = bisect (G, G.source (e), split);
				node leaf = G.new_node ();
				G.new_edge (leaf, G.target (e));
				G.new_edge (leaf, bisector);
				split_fours[e]--;
			} else if (find_splittable_edge (G, split, G.target(e), G.source (e))) {
				node bisector = bisect (G, G.target (e), split);
				node leaf = G.new_node ();
				G.new_edge (leaf, G.source (e));
				G.new_edge (leaf, bisector);
				split_fours[e]--;
			} else {
cout << "COULD NOT FIND SPLITTABLE\n";
				continue;
			}
		}
	}
}

void do_pivot (graph &G, const node &pivot, node &nonpivot, edge &active) {
	node new_nonpivot = G.new_node ();
	if (!degrees.empty ()) {
		if ((dp[new_nonpivot] = degrees.pop ()) == 1) {
			D4--;
		}
	}
	pivoted[new_nonpivot] = false;
	added_to[new_nonpivot] = active;
	back_edge[new_nonpivot] =
		G.new_edge (nonpivot, new_nonpivot);
	active = G.new_edge (pivot, new_nonpivot);
	nonpivot = new_nonpivot;
	dp[pivot]--;
	pivoted[pivot] = true;
}

void swap (node &u, node &v) {
	node temp = u;
	u = v;
	v = temp;
}

bool creates_impossible (int dp_pivot, int dp_nonpivot, list<int> new_degrees) {
	// Check for (5+r)^4
	if ((dp_pivot >= 2) &&
	    (dp_nonpivot == dp_pivot) &&
	    (new_degrees.length () == 2) &&
	    (new_degrees.inf (new_degrees.first ()) == dp_pivot) &&
	    (new_degrees.inf (new_degrees.last ()) == dp_pivot)) {
		return true;
	}
	return false;
}

bool ktree (graph &G, degree_sequence &d, int k) {
	int i;
	//int K, D3, D4;
	int n = d.high () - d.low () + 1;
	int m = 2*n - 3;

	if (k != 2) {
		return false;
	}

	G.clear ();

	// Should check for 2-tree graphical here!

	/*********************************************************************/
	/* Compute K, D4, and D3                                             */
	/*********************************************************************/

	K  = 0;
	D4 = 0;
	D3 = 0;
	for (i = d.high (); i >= d.low (); i--) {
		if (d[i] >= 5) {
			K++;
		} else if (d[i] == 4) {
			D4++;
		} else if (d[i] == 3) {
			D3++;
		} else {
			break;
		}
	}

	node leaf;
	node pivot;
	node nonpivot;

	edge active;

	pivoted.init (G, n+10, false); // Should only need n, but helps
	                               // for debugging
	added_to.init (G, n+10, NULL);
	back_edge.init (G, n+10, NULL);
	dp.init (G, n+10, 0);

	leaves.init (G, 3*m, 0);
	single_fours.init (G, n+10, 0);
	double_fours.init (G, n+10, 0);
	split_fours.init (G, 3*m, 0); // Should only need m, but LEDA gets
	                              // confused when edges are deleted
	threes.init (G, n+10, 0);

	degrees.clear ();

	/*********************************************************************/
	/* Check for edge, triangle and fan                                  */
	/*********************************************************************/

	if ((K+D4) <= 1) {
		// Must be an edge, triangle or fan
		pivot    = G.new_node ();
		nonpivot = G.new_node ();
		G.new_edge (pivot, nonpivot);

		if ((d.high () - d.low ()) == 1) {
			// We have an edge, so we're done
			return true;
		}

		leaf = G.new_node ();
		G.new_edge (leaf, pivot);
		G.new_edge (leaf, nonpivot);

		if (d[d.high ()] == 3) {
			D3--;
		}
		if (d[d.high ()] == 4) {
			D4--;
		}

		// Must be a fan
		while (D3 > 0) {
			node new_nonpivot = G.new_node ();
			G.new_edge (new_nonpivot, pivot);
			G.new_edge (new_nonpivot, nonpivot);
			nonpivot = new_nonpivot;
			D3--;
		}
		return true;
	}

	/*********************************************************************/
	/*                                                                   */
	/*********************************************************************/

#if 1
	for (i = 0; i < (K+D4); i++) {
		degrees.append (d[d.high () - i] - 3);
	}
#else
	// Make second-last big
	for (i = 0; i < K-1; i++) {
		degrees.append (d[d.high () - i] - 3);
	}
	if (K >= 1) {
		degrees.push (d[d.high () - (K-1)] - 3);
	}
	for (i = K; i < (K+D4); i++) {
		degrees.append (d[d.high () - i] - 3);
	}
#endif

	// Create first triangle

	pivot    = G.new_node ();
	nonpivot = G.new_node ();
	active   = G.new_edge (pivot, nonpivot);

	pivoted[pivot]    = false;
	pivoted[nonpivot] = false;

	if ((dp[pivot] = degrees.pop ()) == 1) {
		D4--;
	}
	if ((dp[nonpivot] = degrees.pop ()) == 1) {
		D4--;
	}

	add_leaf (G, active);
	// First two vertices only have degree 2 so far
	dp[pivot]++;
	dp[nonpivot]++;

	for (;;) {
		if (dp[pivot] > 0) {
			// Look for (6+r)^2,(5+s)^2
			if ((D3 == 0) &&
			    (D4 == 0) &&
			    (degrees.length () == 2) &&
			    (dp[pivot] == dp[nonpivot]) &&
			    ((dp[pivot] + leaves[active]) >= 4) &&
			    (dp[pivot] > degrees.inf (degrees.first ())) &&
			    (degrees.inf (degrees.first ()) == degrees.inf (degrees.last ())) &&
			    ((G.degree(pivot) + leaves[active]) == 2) &&
			    ((G.degree(nonpivot) + leaves[active]) == 2)) {
				if (leaves[active] > 0) {
					remove_leaf (G, active);
				}
				leaf = G.new_node ();
				add_leaf (G, G.new_edge (leaf, pivot), degrees.pop () + 1);
				G.new_edge (leaf, nonpivot);
				leaf = G.new_node ();
				add_leaf (G, G.new_edge (leaf, nonpivot), degrees.pop () + 1);
				G.new_edge (leaf, pivot);
				dp[pivot] -= 2;
				dp[nonpivot] -= 2;
				continue;
			}

			// Use 3's
			if ((D3 >= 2) && ((dp[pivot] > 1) || (dp[nonpivot] > 0) || (degrees.length () == 0))) {
				add_three (G, pivot);
				continue;
			}

			// Add two 4's
			if ((dp[pivot] >= 2) && (D4 >= 2) && ((dp[pivot] > 2) || (dp[nonpivot] > 0) || (degrees.length () == 2))) {
				list<int> new_degrees = degrees;
				new_degrees.Pop ();
				new_degrees.Pop ();
				if (!creates_impossible (dp[pivot]-2+leaves[active]-1, dp[nonpivot]+leaves[active]-1, new_degrees)) {
					add_double_four (G, pivot);
					continue;
				}
			}

			// Add a single 4
			if ((dp[pivot] >= 2) && (D4 >= 1) && ((dp[pivot] > 2) || (dp[nonpivot] > 0) || (degrees.length () == 1))) {
				list<int> new_degrees = degrees;
				new_degrees.Pop ();
				if (!creates_impossible (dp[pivot]-2+leaves[active]-1, dp[nonpivot]+leaves[active]-1, new_degrees)) {
					add_single_four (G, pivot);
					continue;
				}
			}

			// Do something strange
			if (((!degrees.empty ()) || (D4 > 0)) && (dp[pivot] == 1) && (dp[nonpivot] == 1)) {
				do_pivot (G, nonpivot, pivot, active);
				add_leaf (G, back_edge[pivot]);
				continue;
			}

			// Neither exhausted
			if (dp[nonpivot] > 0) {
				add_leaf (G, active);
				continue;
			}

			// Non-Pivot exhausted
			if (degrees.empty ()) {
				break;
			}
			do_pivot (G, pivot, nonpivot, active);
		} else {
			if (dp[nonpivot] > 0) {

				// Pivot exhausted
				swap (pivot, nonpivot);

			} else {
				// Both exhausted
				// Either we've finished the last vertex,
				// or something is wrong
				break;
			}
		}
	}

	while (dp[pivot] != D3) {
		if (dp[pivot] > D3) {
			// Try splitting a double four into two single fours
			if (double_fours[pivot] > 0) {
				remove_double_four (G, pivot);
				add_single_four (G, pivot);
				add_single_four (G, pivot);
			}
			// Try pulling double fours down
			else if (double_fours[nonpivot] > 0) {
				remove_double_four (G, nonpivot);
				add_single_four (G, nonpivot);
				add_single_four (G, pivot);
			}
			// Try pulling a single four down
			else if ((single_fours[nonpivot] > 0) && (split_fours[active] == 0)) {
				remove_single_four (G, nonpivot);
				add_split_four (G, active);
				add_leaf (G, active);
			}
			// Try pulling threes down
			else if (threes[nonpivot] > 0) {
				remove_three (G, nonpivot);
				add_three (G, pivot);
				add_leaf (G, active);
			}
			// Try pulling leaves down
			else if ((added_to[pivot] != NULL) && (leaves[added_to[pivot]] > 0)) {
				if ((!is_adjacent (G, pivot, G.source (added_to[pivot]))) ||
				    (!is_adjacent (G, pivot, G.target (added_to[pivot])))) {
					cout << "SOMETHING WRONG!!!!!!!!!!!!!!!!!!!!!!!!\n";
					return false;
				}
				remove_leaf (G, added_to[pivot]);
				add_leaf (G, pivot, G.source (added_to[pivot]));
				add_leaf (G, pivot, G.target (added_to[pivot]));
			} else {
				break;
			}
		} else if (D3 > dp[pivot]) {
			// Try pulling single fours down
			if (single_fours[nonpivot] > 0) {
				remove_single_four (G, nonpivot);
				add_three (G, nonpivot);
				add_three (G, nonpivot);
				//degrees.append (4);
				do_pivot (G, pivot, nonpivot, active);
				add_three (G, nonpivot); // Nonpivot becomes a 4
			} else {
				break;
			}
		}
	}

	// Add remaining 3's if necessary
	while ((D3 > 0) && (dp[pivot] > 0)) {
		add_three (G, pivot);
	}

	// Add last leaf if necessary
	if (dp[pivot] >= 0) {
		add_leaf (G, active);
	}
	dp[pivot]++;
	dp[nonpivot]++;

	post_process_leaves (G);
	post_process_threes (G);
	post_process_fours (G);

#if 0
cout << "D4 " << D4 << "\n";
cout << "D3 " << D3 << "\n";
cout << "dp[pivot] " << dp[pivot] << "\n";
cout << "dp[nonpivot] " << dp[nonpivot] << "\n";
cout << "degrees.length () " << degrees.length () << "\n";
#endif

	return ((degrees.empty ()) &&
	        (dp[pivot] == 0) &&
	        (dp[nonpivot] == 0) &&
	        (D4 == 0) &&
	        (D3 == 0));
}
